import pathlib

import yaml
from yaml.parser import ParserError
from yaml.reader import ReaderError
from yaml.scanner import ScannerError

from .exceptions import InvalidTasksDescription


class YamlLoader:
    def load(self, path: pathlib.Path) -> dict:
        try:
            with path.open("r") as file:
                content = yaml.full_load(file)
        except (ScannerError, ParserError, ReaderError):
            raise InvalidTasksDescription()
        return content



from collections import Counter
from typing import get_args

from pydantic import BaseModel, field_validator

from py_tata.tasks.check_hostname.validation_model import CheckHostnameIn
from py_tata.tasks.check_ping.validation_model import CheckPingIn
from py_tata.tasks.linux_check_ip_forwarding.ipv4 import LinuxCheckIPv4ForwardingIn
from py_tata.tasks.linux_check_ip_forwarding.ipv6 import LinuxCheckIPv6ForwardingIn

_unique_types = (
    CheckHostnameIn,
    LinuxCheckIPv4ForwardingIn,
    LinuxCheckIPv6ForwardingIn,
)
_validation_tasks_types = (
    CheckHostnameIn
    | CheckPingIn
    | LinuxCheckIPv4ForwardingIn
    | LinuxCheckIPv6ForwardingIn
)


class TasksContainer(BaseModel):
    tasks: list[_validation_tasks_types]

    @field_validator("tasks")
    @classmethod
    def check_unique_tasks(cls, v: list):
        type_counts = Counter(type(x) for x in v)
        error_tasks = []
        for unique_type in _unique_types:
            if type_counts[unique_type] < 2:
                continue
            task_name = get_args(unique_type.__annotations__["task"])[0]
            error_tasks.append(f"'{task_name}'")
        if error_tasks:
            raise ValueError(", ".join(error_tasks) + " can be used only once")
        return v



import platform

from py_tata.tasks.task_result import TaskResult


class CheckHostnameTask:
    def __init__(self, target_hostname: str, **kwargs):
        self.target_hostname = target_hostname

    async def check(self) -> TaskResult:
        errors = []
        hostname = platform.node()
        if hostname != self.target_hostname:
            error = (
                f"Incorrect hostname: "
                f"Expected {self.target_hostname} but hostname is {hostname} now"
            )
            errors.append(error)

        return TaskResult(
            description=f"Check hostname is {self.target_hostname}",
            errors=errors,
        )



from typing import Literal

from pydantic import BaseModel, ConfigDict


class CheckHostnameIn(BaseModel):
    task: Literal["check_hostname"]
    target_hostname: str

    model_config = ConfigDict(extra="forbid")



import asyncio
from collections.abc import Coroutine, Iterable
from typing import Protocol

from py_tata.tasks import TaskResult


class TaskProtocol(Protocol):
    async def check(self) -> TaskResult:
        ...


class Runner:
    def __init__(self, tasks: list[TaskProtocol]):
        self.tasks = tasks

    def run(self) -> Iterable[Coroutine[None, None, TaskResult]]:
        asyncio_tasks = []
        for task in self.tasks:
            asyncio_tasks.append(asyncio.create_task(task.check()))

        for coro in asyncio.as_completed(asyncio_tasks):
            yield coro



from py_tata.core.command_executor import asyncio_command_executor

from ..builder import LinuxCheckForwardingTaskBuilder
from .task import LinuxCheckIPv4ForwardingTask

task_builder = LinuxCheckForwardingTaskBuilder(
    command_executor=asyncio_command_executor,
    task_class=LinuxCheckIPv4ForwardingTask,
)



from typing import TypeVar

import pydantic

T = TypeVar("T")


class TaskBuilder:
    def __init__(self, task_class: T):
        self.task_class = task_class

    def build(self, model_in: pydantic.BaseModel) -> T:
        return self.task_class(**model_in.model_dump())



from typing import get_args

from .builder import check_ping_task_builder
from .task import CheckPingTask
from .validation_model import CheckPingIn

TASK_NAME = get_args(CheckPingIn.__annotations__["task"])[0]
builder = check_ping_task_builder
TASK_NAME_TO_BUILDER = {TASK_NAME: builder}



from typing import get_args

from .builder import task_builder
from .task import LinuxCheckIPv4ForwardingTask
from .validation_model import LinuxCheckIPv4ForwardingIn

TASK_NAME = get_args(LinuxCheckIPv4ForwardingIn.__annotations__["task"])[0]
builder = task_builder
TASK_NAME_TO_BUILDER = {TASK_NAME: builder}



from typing import TypeVar

import pydantic

from py_tata.core.builder.task_builder import TaskBuilder
from py_tata.core.command_executor import AsyncioCommandExecutor

T = TypeVar("T")


class LinuxCheckForwardingTaskBuilder(TaskBuilder):
    def __init__(self, command_executor: AsyncioCommandExecutor, task_class: T):
        super().__init__(task_class)
        self.command_executor = command_executor

    def build(
        self,
        model_in: pydantic.BaseModel,
    ) -> T:
        return self.task_class(
            **model_in.model_dump(),
            asyncio_command_executor=self.command_executor,
        )



import asyncio
from pathlib import Path
from typing import Any

import typer
from pydantic import ValidationError
from rich import print as rich_print
from rich.progress import (
    BarColumn,
    MofNCompleteColumn,
    Progress,
    TextColumn,
    TimeElapsedColumn,
)
from typer import Typer

import py_tata.tasks.check_hostname as check_hostname
import py_tata.tasks.check_ping as check_ping
import py_tata.tasks.linux_check_ip_forwarding.ipv4 as linux_check_ipv4_forwarding
import py_tata.tasks.linux_check_ip_forwarding.ipv6 as linux_check_ipv6_forwarding
from py_tata.core.builder.task_builder import TaskBuilder
from py_tata.core.loader import YamlLoader
from py_tata.core.loader.exceptions import InvalidTasksDescription
from py_tata.core.runner import Runner
from py_tata.tasks.validation_model import TasksContainer

_task_name_to_builder: dict[str, TaskBuilder] = {
    **check_hostname.TASK_NAME_TO_BUILDER,
    **check_ping.TASK_NAME_TO_BUILDER,
    **linux_check_ipv4_forwarding.TASK_NAME_TO_BUILDER,
    **linux_check_ipv6_forwarding.TASK_NAME_TO_BUILDER,
}

app = Typer()


@app.command()
def main(tasks_path: Path):
    asyncio.run(async_main(tasks_path))


async def async_main(tasks_path: Path):
    rich_print(f"[cyan]Collecting tasks...[/]")
    raw_content = __load_yaml(tasks_path)
    tasks_container = __build_tasks_container(raw_content)
    tasks = __build_tasks(tasks_container)
    rich_print(f"[bold cyan]{len(tasks_container.tasks)} tasks collected[/]")
    runner = Runner(tasks)

    progress_bar = Progress(
        TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
        BarColumn(),
        MofNCompleteColumn(),
        TextColumn("•"),
        TimeElapsedColumn(),
    )
    with progress_bar as progress:
        task = progress.add_task(description="", total=len(tasks))
        for index, completed_coro in enumerate(runner.run()):
            result = await completed_coro
            if not result.errors:
                rich_print(f"[green][+] Task {result.description}[/]")
            else:
                rich_print(
                    f"[red][-] Task {result.description}\t| {','.join(result.errors)}[/]"
                )
            progress.update(task, completed=index + 1)


def __load_yaml(tasks_path: Path) -> dict:
    try:
        raw_tasks_container = YamlLoader().load(tasks_path)
    except InvalidTasksDescription:
        rich_print(f"[red]Invalid yaml: {tasks_path}[/]")
        raise typer.Exit(1)
    return raw_tasks_container


def __build_tasks_container(raw_yaml: Any) -> TasksContainer:
    try:
        tasks_container = TasksContainer.model_validate(raw_yaml)
    except ValidationError as e:
        rich_print("[bold red]Validation error:[/]")
        rich_print(e)
        raise typer.Exit(1)
    return tasks_container


def __build_tasks(tasks_container: TasksContainer) -> list:
    return [
        _task_name_to_builder[task_in.task].build(task_in)
        for task_in in tasks_container.tasks
    ]


if __name__ == "__main__":
    app()



from py_tata.core.command_executor import asyncio_command_executor

from ..builder import LinuxCheckForwardingTaskBuilder
from .task import LinuxCheckIPv6ForwardingTask

task_builder = LinuxCheckForwardingTaskBuilder(
    command_executor=asyncio_command_executor,
    task_class=LinuxCheckIPv6ForwardingTask,
)



from ..task_abc import LinuxCheckIpForwardingTask


class LinuxCheckIPv4ForwardingTask(LinuxCheckIpForwardingTask):
    @property
    def status_check_command(self) -> tuple[str, ...]:
        return "sysctl", "-n", "net.ipv4.ip_forward"

    @property
    def ip_version(self) -> str:
        return "IPv4"



from icmplib import async_ping

from py_tata.tasks import TaskResult


class CheckPingTask:
    def __init__(self, target_ip: str, **kwargs):
        self.target_ip = target_ip

    async def check(self) -> TaskResult:
        host = await async_ping(self.target_ip, count=3, privileged=False)
        errors = []
        if not host.is_alive:
            errors.append(f"Cannot ping host with ip={self.target_ip}")
        return TaskResult(
            description=f"Check ping for ip={self.target_ip}",
            errors=errors,
        )



import abc
import platform
from abc import ABC, abstractmethod

from py_tata.core.command_executor import AsyncioCommandExecutor
from py_tata.tasks.task_result import TaskResult


class LinuxCheckIpForwardingTask(ABC):
    def __init__(
        self,
        asyncio_command_executor: AsyncioCommandExecutor,
        enabled: bool,
        **kwargs,
    ):
        self.asyncio_command_executor = asyncio_command_executor
        self.enabled = enabled

    @property
    @abstractmethod
    def ip_version(self) -> str:
        ...

    @property
    @abstractmethod
    def status_check_command(self) -> tuple[str, ...]:
        ...

    @property
    def task_description(self):
        return f"Check {self.ip_version} forwarding is {'enabled' if self.enabled else 'disabled'}"

    async def check(self):
        if platform.system() != "Linux":
            return TaskResult(
                description=self.task_description,
                errors=["Not supported system. Only Linux support"],
            )

        errors = []
        try:
            command = self.status_check_command
            command_result = await self.asyncio_command_executor.execute(*command)
            if command_result.return_code != 0:
                errors.append("Failed to check configuration with sysctl")
            if self.enabled and int(command_result.stdout) != 1:
                errors.append(f"{self.ip_version} forwarding disabled")
            if not self.enabled and int(command_result.stdout) != 0:
                errors.append(f"{self.ip_version} forwarding enabled")

        except Exception as e:
            errors.append(str(e))

        return TaskResult(description=self.task_description, errors=errors)



import asyncio
import dataclasses


@dataclasses.dataclass(frozen=True, slots=True)
class CommandResult:
    return_code: int
    stdout: str
    stderr: str


class AsyncioCommandExecutor:
    async def execute(self, *command):
        process = await asyncio.create_subprocess_exec(
            *command,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        stdout, stderr = await process.communicate()

        return CommandResult(
            return_code=process.returncode,
            stdout=stdout.decode().strip(),
            stderr=stderr.decode().strip(),
        )
